#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Adsorbate procreation operators that adds an adsorbate to the surface of 
a particle or given structure."""
from ..settings import (adsorbate_elements, 
                        site_heights)
from ..utilities import (custom_warning, 
                         is_list_or_tuple, 
                         atoms_too_close_after_addition)
from ..adsorption_sites import (ClusterAdsorptionSites, 
                                SlabAdsorptionSites)
from ..adsorbate_coverage import (ClusterAdsorbateCoverage, 
                                  SlabAdsorbateCoverage)
from ..build.action import (add_adsorbate_to_site, 
                            remove_adsorbate_from_site, 
                            remove_adsorbates_from_sites)
from ase.ga.offspring_creator import OffspringCreator
from ase.formula import Formula
from ase import Atoms
from operator import attrgetter
from itertools import chain
from copy import deepcopy
import numpy as np
import warnings
import random
warnings.formatwarning = custom_warning


class AdsorbateOperator(OffspringCreator):
    """Base class for all operators that add, move or remove adsorbates.

    Don't use this operator directly!"""

    def __init__(self, adsorbate_species, species_probabilities=None, num_muts=1):        
        OffspringCreator.__init__(self, num_muts=num_muts)

        self.adsorbate_species = adsorbate_species if is_list_or_tuple(
                                 adsorbate_species) else [adsorbate_species]
        self.species_probabilities = species_probabilities
        if self.species_probabilities is not None:
            assert len(self.species_probabilities.keys()) == len(self.adsorbate_species)

        self.descriptor = 'AdsorbateOperator'

    @classmethod
    def initialize_individual(cls, parent, indi=None):
        indi = OffspringCreator.initialize_individual(parent, indi=indi)
        if 'unrelaxed_adsorbates' in parent.info['data']:
            unrelaxed = list(parent.info['data']['unrelaxed_adsorbates'])
        else:
            unrelaxed = []
        indi.info['data']['unrelaxed_adsorbates'] = unrelaxed
        
        return indi
        
    def get_new_individual(self, parents):
        raise NotImplementedError

    def add_adsorbate(self, atoms, hetero_site_list, heights, 
                      adsorbate_species=None, 
                      min_adsorbate_distance=2., 
                      tilt_angle=0.):
        """Adds the adsorbate in self.adsorbate to the supplied atoms
        object at the first free site in the specified site_list. A site
        is free if no other adsorbates can be found in a sphere of radius
        min_adsorbate_distance around the chosen site.

        Parameters
        ----------
        atoms : ase.Atoms object
            The atoms object that the adsorbate will be added to.

        hetero_site_list : list
            A list of dictionaries, each dictionary contains site 
            information given by acat.adsorbate_coverage module.

        heights : dict
            A dictionary that contains the adsorbate height for each site 
            type.

        adsorbate_species : str or list of strs, default None
            One or a list of adsorbate species to be added to the surface.
            Use self.adsorbate_species if not specified.

        min_adsorbate_distance : float, default 2.
            The radius of the sphere inside which no other adsorbates 
            should be found.
        
        tilt_angle : float, default 0.
            Tilt the adsorbate with an angle (in degrees) relative to the 
            surface normal.

        """

        if adsorbate_species is None:
            adsorbate_species = self.adsorbate_species
        elif not is_list_or_tuple(adsorbate_species): 
            adsorbate_species = [adsorbate_species]
        i = 0
        too_close = True
        while too_close:
            if i >= len(hetero_site_list):
                return False

            site = hetero_site_list[i]
            if site['occupied']:
                i += 1
                continue

            # Allow only single-atom species to enter subsurf 6fold sites
            this_site = site['site']
            if this_site == '6fold':
                adsorbate_species = [s for s in adsorbate_species if len(s) == 1]

            # Add a random adsorbate to the correct position
            if self.species_probabilities is None:
                adsorbate = random.choice(adsorbate_species)
            else:
                probs = [self.species_probabilities[a] for a in adsorbate_species]
                adsorbate = random.choices(k=1, population=adsorbate_species,
                                           weights=probs)[0]                   
            height = heights[this_site]
            add_adsorbate_to_site(atoms, adsorbate, site, height, 
                                  tilt_angle=tilt_angle)

            nads = len(list(Formula(adsorbate)))
            ads_atoms = atoms[[a.index for a in atoms if 
                               a.symbol in adsorbate_elements]]
            if atoms_too_close_after_addition(ads_atoms, nads, 
            cutoff=min_adsorbate_distance):
                atoms = atoms[:-nads]
                i += 1
                continue    

            too_close = False 

        # Setting the indices of the unrelaxed adsorbates for the cut-
        # relax-paste function to be executed in the calculation script.
        # There it should also reset the parameter to [], to indicate
        # that the adsorbates have been relaxed.
        ads_indices = sorted([len(atoms) - k - 1 for k in range(nads)])
        
        if 'unrelaxed_adsorbates' not in atoms.info['data']:
            atoms.info['data']['unrelaxed_adsorbates'] = []
        atoms.info['data']['unrelaxed_adsorbates'].append(ads_indices)
        
        return atoms

    def remove_adsorbate(self, atoms, hetero_site_list, return_site_index=False):                          
        """Removes an adsorbate from the atoms object at the first occupied
        site in hetero_site_list. If no adsorbates can be found, one will be
        added instead.

        Parameters
        ----------
        atoms : ase.Atoms object
            The atoms object that the adsorbate will be added to

        hetero_site_list : list
            A list of dictionaries, each dictionary contains site 
            information given by acat.adsorbate_coverage module.

        return_site_index : bool, default False
            Whether to return the site index of the hetero_site_list instead
            of the site. Useful for moving or replacing adsorbate.

        """

        i = 0
        occupied = False
        while not occupied:
            if i >= len(hetero_site_list):
                if return_site_index:
                    return False
                warnings.warn('Removal not possible, will add instead')
                return self.add_adsorbate(atoms, hetero_site_list, site_heights)

            site = hetero_site_list[i]
            if not site['occupied']:
                i += 1
                continue
            # Remove adsorbate from the correct position. Remove the fragment
            # with priority if it is one of the given adsorbate species
            rm_frag = (site['fragment'] in self.adsorbate_species) 
            remove_adsorbate_from_site(atoms, site, remove_fragment=rm_frag)

            if return_site_index:
                return i
            occupied = True

        return atoms

    def get_all_adsorbate_indices(self, atoms):
        from asap3 import FullNeighborList

        ac = atoms.copy()
        ads_ind = [a.index for a in ac if 
                   a.symbol in adsorbate_elements]
        mbl = 1.5  # max_bond_length
        nl = FullNeighborList(rCut=mbl / 2., atoms=ac)

        adsorbates = []
        while len(ads_ind) != 0:
            i = int(ads_ind[0])
            mol_ind = self._get_indices_in_adsorbate(ac, nl, i)
            for ind in mol_ind:
                if int(ind) in ads_ind:
                    ads_ind.remove(int(ind))
            adsorbates.append(sorted(mol_ind))
        return adsorbates

    def get_adsorbate_indices(self, atoms, position):
        """Returns the indices of the adsorbate at the supplied position."""
        dmin = 1000.
        for a in atoms:
            if a.symbol in adsorbate_elements:
                d = np.linalg.norm(a.position - position)
                if d < dmin:
                    dmin = d
                    ind = a.index

        for ads in self.get_all_adsorbate_indices(atoms):
            if ind in ads:
                return ads[:]
        
    def _get_indices_in_adsorbate(self, atoms, neighborlist,
                                  index, molecule_indices=None):
        """Internal recursive function that help
        determine adsorbate indices."""
        if molecule_indices is None:
            molecule_indices = []
        mi = molecule_indices
        nl = neighborlist
        mi.append(index)
       # neighbors, _ = nl.get_neighbors(index)
        neighbors = nl[index]

        for n in neighbors:
            if int(n) not in mi:
                if atoms[int(n)].symbol in adsorbate_elements:
                    mi = self._get_indices_in_adsorbate(atoms, nl, n, mi)
        return mi


class AddAdsorbate(AdsorbateOperator):
    """
    Use this operator to add adsorbates to the surface.
    The surface is allowed to change during the algorithm run.

    There is no limit of adsorbate species. You can either provide one
    species or a list of species.

    Site and surface preference can be supplied. If both are supplied site
    will be considered first.

    Supplying a tilt angle will tilt the adsorbate with an angle relative
    to the standard perpendicular to the surface.

    The operator is generalized for both periodic and non-periodic systems 
    (distinguished by atoms.pbc).

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be added to the surface.

    species_probabilities : dict, default None
        A dictionary that contains keys of each adsorbate species and 
        values of their probabilities of adding onto the surface.
        Adding adsorbate species with equal probability if not specified.

    heights : dict, default acat.settings.site_heights
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site 
        type is not specified.

    min_adsorbate_distance : float, default 2.
        The radius of the sphere inside which no other adsorbates 
        should be found.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        genetic algorithm. Make sure all the operators used with this
        operator preserve the indexing of the atoms.

    site_preference : str of list of strs, defualt None
        The site type(s) that has higher priority to attach adsorbates.

    surface_preference : str, default None
        The surface type that has higher priority to attach adsorbates.
        Please only use this for nanoparticles.

    max_coverage : float, default None
        The maximum allowed coverage on the surface. Coverage is defined
        as (number of surface occupied sites / number of surface atoms).
        The maximum coverage is solely governed by min_adsorbate_distance 
        if max_coverage is not specified.

    tilt_angle : float, default 0.
        Tilt the adsorbate with an angle (in degrees) relative to the 
        surface normal.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    num_muts : int, default 1
        The number of times to perform this operation.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 species_probabilities=None,
                 heights=site_heights,
                 min_adsorbate_distance=2.,
                 adsorption_sites=None,
                 site_preference=None,
                 surface_preference=None,
                 max_coverage=None,
                 tilt_angle=None,
                 subtract_heights=False,
                 num_muts=1,
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species,
                                   species_probabilities,
                                   num_muts=num_muts)
        self.descriptor = 'AddAdsorbate'

        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.min_adsorbate_distance = min_adsorbate_distance
        self.adsorption_sites = adsorption_sites
        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)

        self.site_preference = site_preference
        if site_preference is not None:
            if not is_list_or_tuple(site_preference):
                self.site_preference = [site_preference]
        self.surface_preference = surface_preference

        self.max_coverage = max_coverage        
        self.tilt_angle = tilt_angle or 0.
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 1
        self.dmax = dmax

    def get_new_individual(self, parents):
        """Returns the new individual as an atoms object."""
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for atom in f:
            indi.append(atom)

        if True in indi.pbc:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = SlabAdsorptionSites(indi, **self.kwargs)
            sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = ClusterAdsorptionSites(indi, **self.kwargs)
            sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)
        ads_sites = sac.hetero_site_list

        for _ in range(self.num_muts):
            # Make sure the coverage does not exceed maximum coverage
            if self.max_coverage is not None:
                if sac.coverage >= self.max_coverage:
                    break

            random.shuffle(ads_sites)
            if self.surface_preference is not None:
                def func(x):
                    return x['surface'] == self.surface_preference
                ads_sites.sort(key=func, reverse=True)

            if self.site_preference is not None:
                def func(x):
                    return x['site'] in self.site_preference
                ads_sites.sort(key=func, reverse=True)

            nindi = self.add_adsorbate(indi, ads_sites, self.heights,
                                       self.adsorbate_species,
                                       self.min_adsorbate_distance,
                                       tilt_angle=self.tilt_angle)
            if not nindi:
                break

            indi = nindi
            if True in indi.pbc:
                sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                            self.subtract_heights, dmax=self.dmax)
            else:
                sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                               self.subtract_heights, dmax=self.dmax)
            ads_sites = sac.hetero_site_list                          
        
        indi.info['data']['adsorbates'] = [t[0] for t in 
            sac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ': {0}'.format(f.info['confid']))


class RemoveAdsorbate(AdsorbateOperator):
    """This operator removes an adsorbate from the surface. It works
    exactly (but doing the opposite) as the AddAdsorbate operator.

    The operator is generalized for both periodic and non-periodic systems 
    (distinguished by atoms.pbc).

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be removed from the surface.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        genetic algorithm. Make sure all the operators used with this
        operator preserve the indexing of the atoms.

    site_preference : str or list of strs, defualt None
        The site type(s) that has higher priority to detach adsorbates.

    surface_preference : str, default None
        The surface type that has higher priority to detach adsorbates.
        Please only use this for nanoparticles.

    min_coverage : float, default None
        The minimum allowed coverage on the surface. Coverage is defined
        as (number of surface occupied sites / number of surface atoms).
        The minimum coverage is 0 if min_coverage is not specified.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    num_muts : int, default 1
        The number of times to perform this operation.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 adsorption_sites=None,
                 site_preference=None,
                 surface_preference=None,
                 min_coverage=None,
                 subtract_heights=False,
                 num_muts=1,
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species,
                                   num_muts=num_muts)
        self.descriptor = 'RemoveAdsorbate'

        self.adsorption_sites = adsorption_sites
        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)

        self.site_preference = site_preference
        if site_preference is not None:
            if not is_list_or_tuple(site_preference):
                self.site_preference = [site_preference]

        self.min_coverage = min_coverage
        self.surface_preference = surface_preference
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 1
        self.dmax = dmax

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for atom in f:
            indi.append(atom)

        if True in indi.pbc:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = SlabAdsorptionSites(indi, **self.kwargs)
            sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = ClusterAdsorptionSites(indi, **self.kwargs)
            sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)
        ads_sites = sac.hetero_site_list

        for _ in range(self.num_muts):
            # Make sure the coverage is not lower than the minimum coverage
            if self.min_coverage is not None:
                if sac.coverage <= self.min_coverage:
                    break

            random.shuffle(ads_sites)
            if self.surface_preference is not None:
                def func(x):
                    return x['surface'] == self.surface_preference
                ads_sites.sort(key=func, reverse=True)

            if self.site_preference is not None:
                def func(x):
                    return x['site'] in self.site_preference
                ads_sites.sort(key=func, reverse=True)

            indi = self.remove_adsorbate(indi, ads_sites)

            # Make sure there are still adsorbates to remove
            if not [a for a in indi if a.symbol in adsorbate_elements]:
                break
            if True in indi.pbc:
                sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                            self.subtract_heights, dmax=self.dmax)
            else:
                sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                               self.subtract_heights, dmax=self.dmax)
            ads_sites = sac.hetero_site_list                          

        indi.info['data']['adsorbates'] = [t[0] for t in 
            sac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ': {0}'.format(f.info['confid']))


class MoveAdsorbate(AdsorbateOperator):                                           
    """This operator removes an adsorbate from the surface and adds it
    again to a different site, i.e. effectively moving the adsorbate.

    The operator is generalized for both periodic and non-periodic systems 
    (distinguished by atoms.pbc).

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be added to the surface.

    heights : dict, default acat.settings.site_heights
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site 
        type is not specified.

    min_adsorbate_distance : float, default 2.
        The radius of the sphere inside which no other adsorbates 
        should be found.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        genetic algorithm. Make sure all the operators used with this
        operator preserve the indexing of the atoms.

    site_preference_from : str or list of strs, defualt None
        The site type(s) that has higher priority to detach adsorbates.

    surface_preference_from : str, default None
        The surface type that has higher priority to detach adsorbates.
        Please only use this for nanoparticles.

    site_preference_to : str or list of strs, defualt None
        The site type(s) that has higher priority to attach adsorbates.

    surface_preference_to : str, default None
        The surface type that has higher priority to attach adsorbates.
        Please only use this for nanoparticles.

    tilt_angle : float, default 0.
        Tilt the adsorbate with an angle (in degrees) relative to the 
        surface normal.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    num_muts : int, default 1
        The number of times to perform this operation.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 heights=site_heights,
                 min_adsorbate_distance=2.,
                 adsorption_sites=None,
                 site_preference_from=None,
                 surface_preference_from=None,
                 site_preference_to=None,
                 surface_preference_to=None,
                 tilt_angle=None,
                 subtract_heights=False,
                 num_muts=1,
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species,
                                   num_muts=num_muts)
        self.descriptor = 'MoveAdsorbate'

        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.min_adsorbate_distance = min_adsorbate_distance
        self.adsorption_sites = adsorption_sites
        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)

        self.site_preference_from = site_preference_from
        if site_preference_from is not None:
            if not is_list_or_tuple(site_preference_from):
                self.site_preference_from = [site_preference_from]
        self.surface_preference_from = surface_preference_from
        self.site_preference_to = site_preference_to
        if site_preference_to is not None:
            if not is_list_or_tuple(site_preference_to):
                self.site_preference_to = [site_preference_to]
        self.surface_preference_to = surface_preference_to
        self.tilt_angle = tilt_angle or 0.
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 1
        self.dmax = dmax

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for atom in f:
            indi.append(atom)

        if True in indi.pbc:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = SlabAdsorptionSites(indi, **self.kwargs)
            sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = ClusterAdsorptionSites(indi, **self.kwargs) 
            sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)
        ads_sites = sac.hetero_site_list

        for _ in range(self.num_muts):
            random.shuffle(ads_sites)
            if self.surface_preference_from is not None:
                def func(x):
                    return x['surface'] == self.surface_preference_from
                ads_sites.sort(key=func, reverse=True)

            if self.site_preference_from is not None:
                def func(x):
                    return x['site'] in self.site_preference_from
                ads_sites.sort(key=func, reverse=True)

            removed = self.remove_adsorbate(indi, ads_sites,
                                            return_site_index=True)
            if removed is False:
                removed_species = random.choice(self.adsorbate_species)
            else:
                removed_species = ads_sites[removed]['fragment']            
            random.shuffle(ads_sites)

            if self.surface_preference_to is not None:
                def func(x):
                    return x['surface'] == self.surface_preference_to
                ads_sites.sort(key=func, reverse=True)

            if self.site_preference_to is not None:
                def func(x):
                    return x['site'] in self.site_preference_to
                ads_sites.sort(key=func, reverse=True)

            nindi = self.add_adsorbate(indi, ads_sites, 
                                       self.heights, removed_species,
                                       self.min_adsorbate_distance,
                                       tilt_angle=self.tilt_angle)
            if not nindi:
                break

            indi = nindi
            if True in indi.pbc:
                sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                            self.subtract_heights, dmax=self.dmax)
            else:
                sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                               self.subtract_heights, dmax=self.dmax)
            ads_sites = sac.hetero_site_list                          

        indi.info['data']['adsorbates'] = [t[0] for t in 
            sac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ': {0}'.format(f.info['confid']))


class ReplaceAdsorbate(AdsorbateOperator):                                           
    """This operator removes an adsorbate from the surface and adds another
    species to the same site, i.e. effectively replacing the adsorbate.

    The operator is generalized for both periodic and non-periodic systems 
    (distinguished by atoms.pbc).

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be added to the surface.

    species_probabilities : dict, default None
        A dictionary that contains keys of each adsorbate species and 
        values of their probabilities of replacing an adsorbate on the 
        surface. Choosing adsorbate species with equal probability if 
        not specified.

    heights : dict, default acat.settings.site_heights
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site 
        type is not specified.

    min_adsorbate_distance : float, default 2.
        The radius of the sphere inside which no other adsorbates 
        should be found.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        genetic algorithm. Make sure all the operators used with this
        operator preserve the indexing of the atoms.

    site_preference : str or list of strs, defualt None
        The site type(s) that has higher priority to replace adsorbates.

    surface_preference : str, default None
        The surface type that has higher priority to replace adsorbates.
        Please only use this for nanoparticles.

    tilt_angle : float, default 0.
        Tilt the adsorbate with an angle (in degrees) relative to the 
        surface normal.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    num_muts : int, default 1
        The number of times to perform this operation.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 species_probabilities=None,
                 heights=site_heights,
                 min_adsorbate_distance=2.,
                 adsorption_sites=None,
                 site_preference=None,
                 surface_preference=None,
                 tilt_angle=None,
                 subtract_heights=False,
                 num_muts=1,
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species,
                                   species_probabilities,
                                   num_muts=num_muts)
        self.descriptor = 'ReplaceAdsorbate'

        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.min_adsorbate_distance = min_adsorbate_distance
        self.adsorption_sites = adsorption_sites
        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)

        self.site_preference = site_preference
        if site_preference is not None:
            if not is_list_or_tuple(site_preference):
                self.site_preference = [site_preference]
        self.surface_preference = surface_preference

        self.tilt_angle = tilt_angle or 0.
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 1
        self.dmax = dmax

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for atom in f:
            indi.append(atom)

        if True in indi.pbc:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = SlabAdsorptionSites(indi, **self.kwargs)
            sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = ClusterAdsorptionSites(indi, **self.kwargs) 
            sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)
        ads_sites = sac.hetero_site_list

        for _ in range(self.num_muts):
            random.shuffle(ads_sites)
            if self.surface_preference is not None:
                def func(x):
                    return x['surface'] == self.surface_preference
                ads_sites.sort(key=func, reverse=True)

            if self.site_preference is not None:
                def func(x):
                    return x['site'] in self.site_preference
                ads_sites.sort(key=func, reverse=True)

            removed = self.remove_adsorbate(indi, ads_sites,
                                            return_site_index=True)
            if removed is False:
                removed = random.choice(range(len(ads_sites)))
                removed_species = 'x'
            else:
                removed_species = ads_sites[removed]['fragment']
            ads_sites[removed]['occupied'] = 0
            other_species = [s for s in self.adsorbate_species if
                             s != removed_species]

            nindi = self.add_adsorbate(indi, [ads_sites[removed]], 
                                       self.heights, other_species,
                                       self.min_adsorbate_distance,
                                       tilt_angle=self.tilt_angle)
            if not nindi:
                break

            indi = nindi
            if True in indi.pbc:
                sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                            self.subtract_heights, dmax=self.dmax)
            else:
                sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                               self.subtract_heights, dmax=self.dmax)
            ads_sites = sac.hetero_site_list                          

        indi.info['data']['adsorbates'] = [t[0] for t in 
            sac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ': {0}'.format(f.info['confid']))


class ReplaceAdsorbateSpecies(AdsorbateOperator):                                           
    """This operator replace all adsorbates of a certain species with 
    another species at the same sites. Return None if there is no
    adsorbate present on the surface.

    The operator is generalized for both periodic and non-periodic systems 
    (distinguished by atoms.pbc).

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be added to the surface.

    species_probabilities : dict, default None                         
        A dictionary that contains keys of each adsorbate species and 
        values of their probabilities of replacing an adsorbate on the 
        surface. Choosing adsorbate species with equal probability if 
        not specified.

    replace_vacancy : bool, default False
        Whether to allow replacing adsorbates with vacancies, i.e., 
        effectively removing all adsorbates of a certain species.
        Note that if you want to specify species_probabilties, you
        then need to also provide the probability for vacancy
        replacement using the keyword 'vacancy'.

    heights : dict, default acat.settings.site_heights
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site 
        type is not specified.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        genetic algorithm. Make sure all the operators used with this
        operator preserve the indexing of the atoms.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    tilt_angle : float, default 0.
        Tilt the adsorbate with an angle (in degrees) relative to the 
        surface normal.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 replace_vacancy=False,
                 species_probabilities=None,
                 heights=site_heights,
                 adsorption_sites=None,
                 tilt_angle=None,
                 subtract_heights=False, 
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species,
                                   species_probabilities)
        self.descriptor = 'ReplaceAdsorbateSpecies'

        assert len(set(self.adsorbate_species)) > 1
        self.replace_vacancy = replace_vacancy
        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.adsorption_sites = adsorption_sites
        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)

        self.tilt_angle = tilt_angle or 0.
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 1
        self.dmax = dmax

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]
        if 'groups' in f.info['data']:
            indi.info['data']['groups'] = f.info['data']['groups']
        for atom in f:
            indi.append(atom)

        if True in indi.pbc:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = SlabAdsorptionSites(indi, **self.kwargs)
            sac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(indi)
            else:
                sas = ClusterAdsorptionSites(indi, **self.kwargs) 
            sac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)

        ads_sites = sac.hetero_site_list
        specs = [t[0] for t in sac.get_adsorbates(self.adsorbate_species)]
        # Return None if no adsorbate is present
        if not specs:
            return None, '{0} not possible in {1}'.format(self.descriptor,
                                                          f.info['confid'])
        spec = random.choice(specs)
        all_specs = self.adsorbate_species.copy()
        if self.replace_vacancy:
            if 'vacancy' not in all_specs:
                all_specs.append('vacancy')

        other_specs = [sp for sp in all_specs if sp != spec]
        if self.species_probabilities is None:
            to_spec = random.choice(other_specs)
        else:
            probs = []
            for osp in other_specs:
                if (osp == 'vacancy') and ('vacancy' not in self.species_probabilities):
                    vs = self.species_probabilities.values()
                    prob = sum(vs) / len(vs)
                else:
                    prob = self.species_probabilities[osp]
                probs.append(prob)
            to_spec = random.choices(k=1, population=other_specs, weights=probs)[0]  

        rmsites = []
        rmsis = []
        rmfrags = True
        for st in ads_sites:
            if st['fragment'] in self.adsorbate_species:
                if st['fragment'] == spec:
                    rmsites.append(st) 
                    rmsis.append(st['indices'])
            else:
                if st['adsorbate'] == spec:
                    rmsites.append(st)
                    rmsis.append(st['indices'])
                    rmfrags = False
        remove_adsorbates_from_sites(indi, sites=rmsites, remove_fragments=rmfrags)

        if to_spec != 'vacancy':
            for st in ads_sites:
                if st['indices'] in rmsis:
                    height = self.heights[st['site']]
#                    height = st['bond_length']
                    add_adsorbate_to_site(indi, to_spec, st, height, tilt_angle=self.tilt_angle) 
        
        if True in indi.pbc:
            nsac = SlabAdsorbateCoverage(indi, sas, subtract_heights=
                                         self.subtract_heights, dmax=self.dmax)
        else:
            nsac = ClusterAdsorbateCoverage(indi, sas, subtract_heights=
                                            self.subtract_heights, dmax=self.dmax)

        indi.info['data']['adsorbates'] = [t[0] for t in 
            nsac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ': {0}'.format(f.info['confid']))

        
class CutSpliceCrossoverWithAdsorbates(AdsorbateOperator):
    """Crossover that cuts two particles with adsorbates through a plane 
    in space and merges two halfes from different particles together 
    (only returns one of them). The indexing of the atoms is not 
    preserved. Please only use this operator if the particle is allowed 
    to change shape.

    It keeps the correct composition by randomly assigning elements in
    the new particle. If some of the atoms in the two particle halves
    are too close, the halves are moved away from each other perpendicular
    to the cutting plane.

    The complexity of crossover with adsorbates makes this operator not 
    robust enough. The adsorption site identification will fail once the
    nanoparticle shape becomes too irregular after crossover. 

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be added to the surface.

    blmin : dict
        Dictionary of minimum distance between atomic numbers.
        e.g. {(28,29): 1.5}
    
    heights : dict, default acat.settings.site_heights
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site 
        type is not specified.

    keep_composition : bool, default True
        Should the composition be the same as in the parents.

    fix_coverage : bool, default False 
        Should the adsorbate coverage be the same as in the parents.

    min_adsorbate_distance : float, default 2.
        The radius of the sphere inside which no other adsorbates 
        should be found.

    rotate_vectors : list, default None
        A list of vectors that the part of the structure that is cut
        is able to rotate around, the size of rotation is set in
        rotate_angles. Default None meaning no rotation is performed.

    rotate_angles : list, default None
        A list of angles that the structure cut can be rotated. The 
        vector being rotated around is set in rotate_vectors. Default 
        None meaning no rotation is performed.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species, blmin, 
                 heights=site_heights,
                 keep_composition=True,
                 fix_coverage=False, 
                 min_adsorbate_distance=2.,
                 rotate_vectors=None, 
                 rotate_angles=None,
                 subtract_heights=False,
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species)
        self.descriptor = 'CutSpliceCrossoverWithAdsorbates'

        self.blmin = blmin
        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.keep_composition = keep_composition
        self.fix_coverage = fix_coverage
        self.min_adsorbate_distance = min_adsorbate_distance
        self.rvecs = rotate_vectors
        self.rangs = rotate_angles
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 2
        self.dmax = dmax

        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        self.__dict__.update(self.kwargs)
        
    def get_new_individual(self, parents):
        f, m = parents

        if self.fix_coverage:
            # Count number of adsorbates
            adsorbates_in_parents = len(self.get_all_adsorbate_indices(f))
            
        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [i.info['confid'] for i in parents]
        
        fna = self.get_atoms_without_adsorbates(f)
        mna = self.get_atoms_without_adsorbates(m)
        fna_geo_mid = np.average(fna.get_positions(), 0)
        mna_geo_mid = np.average(mna.get_positions(), 0)
        
        if self.rvecs is not None:
            if not isinstance(self.rvecs, list):
                warnings.warn('Rotation vectors are not a list, skipping rotation')
            else:
                vec = random.choice(self.rvecs)
                try:
                    angle = random.choice(self.rangs)
                except TypeError:
                    angle = self.rangs
                f.rotate(angle, vec, center=fna_geo_mid)
                vec = random.choice(self.rvecs)
                try:
                    angle = random.choice(self.rangs)
                except TypeError:
                    angle = self.rangs
                m.rotate(angle, vec, center=mna_geo_mid)
                
        theta = random.random() * 2 * np.pi  # 0,2pi
        phi = random.random() * np.pi  # 0,pi
        e = np.asarray((np.sin(phi) * np.cos(theta),
                        np.sin(theta) * np.sin(phi),
                        np.cos(phi)))
        eps = 0.0001
        
        # Move each particle to origo with their respective geometrical
        # centers, without adsorbates
        common_mid = (fna_geo_mid + mna_geo_mid) / 2.
        f.translate(-common_mid)
        m.translate(-common_mid)
        
        off = 1
        while off != 0:
            fna = self.get_atoms_without_adsorbates(f)
            mna = self.get_atoms_without_adsorbates(m)

            # Get the signed distance to the cutting plane
            # We want one side from f and the other side from m
            fmap = [np.dot(x, e) for x in fna.get_positions()]
            mmap = [-np.dot(x, e) for x in mna.get_positions()]
            ain = sorted([i for i in chain(fmap, mmap) if i > 0],
                         reverse=True)
            aout = sorted([i for i in chain(fmap, mmap) if i < 0],
                          reverse=True)

            off = len(ain) - len(fna)

            # Translating f and m to get the correct number of atoms
            # in the offspring
            if off < 0:
                # too few
                # move f and m away from the plane
                dist = abs(aout[abs(off) - 1]) + eps
                f.translate(e * dist)
                m.translate(-e * dist)
            elif off > 0:
                # too many
                # move f and m towards the plane
                dist = abs(ain[-abs(off)]) + eps
                f.translate(-e * dist)
                m.translate(e * dist)
            eps /= 5.

        fna = self.get_atoms_without_adsorbates(f)
        mna = self.get_atoms_without_adsorbates(m)
        
        # Determine the contributing parts from f and m
        tmpf, tmpm = Atoms(), Atoms()
        for atom in fna:
            if np.dot(atom.position, e) > 0:
                atom.tag = 1
                tmpf.append(atom)
        for atom in mna:
            if np.dot(atom.position, e) < 0:
                atom.tag = 2
                tmpm.append(atom)

        # Place adsorbates from f and m in tmpf and tmpm
        f_ads = self.get_all_adsorbate_indices(f)
        m_ads = self.get_all_adsorbate_indices(m)
        for ads in f_ads:
            if np.dot(f[ads[0]].position, e) > 0:
                for i in ads:
                    f[i].tag = 1
                    tmpf.append(f[i])
        for ads in m_ads:
            pos = m[ads[0]].position
            if np.dot(pos, e) < 0:
                # If the adsorbate will sit too close to another adsorbate
                # (below self.min_adsorbate_distance) do not add it.
                dists = [np.linalg.norm(pos - a.position)
                         for a in tmpf if a.tag == 1]
                for d in dists:
                    if d < self.min_adsorbate_distance:
                        break
                else:
                    for i in ads:
                        m[i].tag = 2
                        tmpm.append(m[i])
                
        tmpfna = self.get_atoms_without_adsorbates(tmpf)
        tmpmna = self.get_atoms_without_adsorbates(tmpm)
                
        # Check that the correct composition is employed
        if self.keep_composition:
            opt_sm = sorted(fna.numbers)
            tmpf_numbers = list(tmpfna.numbers)
            tmpm_numbers = list(tmpmna.numbers)
            cur_sm = sorted(tmpf_numbers + tmpm_numbers)
            # correct_by: dictionary that specifies how many
            # of the atom_numbers should be removed (a negative number)
            # or added (a positive number)
            correct_by = dict([(j, opt_sm.count(j)) for j in set(opt_sm)])
            for n in cur_sm:
                correct_by[n] -= 1
            correct_in = random.choice([tmpf, tmpm])
            to_add, to_rem = [], []
            for num, amount in correct_by.items():
                if amount > 0:
                    to_add.extend([num] * amount)
                elif amount < 0:
                    to_rem.extend([num] * abs(amount))
            for add, rem in zip(to_add, to_rem):
                tbc = [a.index for a in correct_in if a.number == rem]
                if len(tbc) == 0:
                    pass
                ai = random.choice(tbc)
                correct_in[ai].number = add
                
        # Move the contributing apart if any distance is below blmin
        maxl = 0.
        for sv, min_dist in self.get_vectors_below_min_dist(tmpf + tmpm):
            lsv = np.linalg.norm(sv)  # length of shortest vector
            d = [-np.dot(e, sv)] * 2
            d[0] += np.sqrt(np.dot(e, sv)**2 - lsv**2 + min_dist**2)
            d[1] -= np.sqrt(np.dot(e, sv)**2 - lsv**2 + min_dist**2)
            l = sorted([abs(i) for i in d])[0] / 2. + eps
            if l > maxl:
                maxl = l
        tmpf.translate(e * maxl)
        tmpm.translate(-e * maxl)
        
        # Translate particles halves back to the center
        tmpf.translate(common_mid)
        tmpm.translate(common_mid)

        # Put the two parts together
        for atom in chain(tmpf, tmpm):
            indi.append(atom)

        pcas = ClusterAdsorptionSites(indi, **self.kwargs) 
        pcac = ClusterAdsorbateCoverage(indi, pcas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        pads_sites = pcac.hetero_site_list       

        adsi_dict = {}
        for st in pads_sites:
            if st['occupied']:
                si = st['indices']
                adsi_dict[si] = {}
                adsi_dict[si]['height'] = self.heights[st['site']]
                adsi_dict[si]['fragment'] = st['fragment']
                adsi_dict[si]['fragment_indices'] = st['fragment_indices']

        indi = pcas.atoms                
        indi.positions = pcas.ref_atoms.positions
        cas = ClusterAdsorptionSites(indi, **self.kwargs) 

        nori = len(indi) 
        for st in cas.site_list:
            si = st['indices']
            if si in adsi_dict:
                adsorbate = adsi_dict[si]['fragment']
                height = adsi_dict[si]['height']    
                add_adsorbate_to_site(indi, adsorbate, st, height)

                # Make sure no adsorbates too close to each other 
                # after each adsorbate addition
                nads = len(adsi_dict[si]['fragment_indices'])
                if atoms_too_close_after_addition(indi, nads,
                self.min_adsorbate_distance, mic=False):
                    indi = indi[:-nads]                               

        # Add adsorbate if no adsorbate is present
        if len(indi) == nori:
            st = random.choice(cas.site_list)
            ads_spec = random.choice(self.adsorbate_species)
            add_adsorbate_to_site(indi, ads_spec, site=st, 
                                  height=self.heights[st['site']])

        cac = ClusterAdsorbateCoverage(indi, cas, subtract_heights=
                                       self.subtract_heights, dmax=self.dmax)
        ads_sites = cac.hetero_site_list

        if self.fix_coverage:
            # Remove or add adsorbates as needed
            adsorbates_in_child = self.get_all_adsorbate_indices(indi)
            diff = len(adsorbates_in_child) - adsorbates_in_parents
            if diff < 0:
                # Add adsorbates
                for _ in range(abs(diff)):
                    self.add_adsorbate(indi, ads_sites, site_heights,
                                       self.adsorbate_species,
                                       self.min_adsorbate_distance)
            elif diff > 0:
                # Remove adsorbates
                tbr = random.sample(adsorbates_in_child, diff)  # to be removed
                for adsorbate_indices in sorted(tbr, reverse=True):
                    for i in adsorbate_indices[::-1]:
                        indi.pop(i)                
            cac = ClusterAdsorbateCoverage(indi, cas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)

        indi.info['data']['adsorbates'] = [t[0] for t in 
            cac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ': {0} {1}'.format(f.info['confid'],
                                                     m.info['confid']))

    def get_numbers(self, atoms):
        """Returns the atomic numbers of the atoms object
        without adsorbates"""
        ac = atoms.copy()
        del ac[[a.index for a in ac
                if a.symbol in adsorbate_elements]]
        return ac.numbers
        
    def get_atoms_without_adsorbates(self, atoms):
        ac = atoms.copy()
        del ac[[a.index for a in ac
                if a.symbol in adsorbate_elements]]
        return ac
        
    def get_vectors_below_min_dist(self, atoms):
        """Generator function that returns each vector (between atoms)
        that is shorter than the minimum distance for those atom types
        (set during the initialization in blmin)."""
        ap = atoms.get_positions()
        an = atoms.numbers
        for i in range(len(atoms)):
            pos = atoms[i].position
            for j, d in enumerate([np.linalg.norm(k - pos) for k in ap[i:]]):
                if d == 0:
                    continue
                min_dist = self.blmin[tuple(sorted((an[i], an[j + i])))]
                if d < min_dist:
                    yield atoms[i].position - atoms[j + i].position, min_dist


class SimpleCutSpliceCrossoverWithAdsorbates(AdsorbateOperator):
    """Crossover that divides two particles through a plane in space and
    merges the symbols of two halves from different particles with 
    adosrbates together (only returns one of them). The indexing of the 
    atoms is preserved. Please only use this operator with other operators 
    that also preserves the indexing.

    It keeps the correct composition by randomly assigning elements in
    the new particle.

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        One or a list of adsorbate species to be added to the surface.

    heights : dict, default acat.settings.site_heights
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site 
        type is not specified.

    keep_composition : bool, default True
        Boolean that signifies if the composition should be the same 
        as in the parents.

    fix_coverage : bool, default False 
        Should the adsorbate coverage be the same as in the parents.

    min_adsorbate_distance : float, default 2.
        The radius of the sphere inside which no other adsorbates 
        should be found.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object \
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        genetic algorithm. Make sure all the operators used with this
        operator preserve the indexing of the atoms.

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 heights=site_heights,
                 keep_composition=True,
                 fix_coverage=False, 
                 min_adsorbate_distance=2.,
                 adsorption_sites=None, 
                 subtract_heights=False,
                 dmax=2.5, **kwargs):
        AdsorbateOperator.__init__(self, adsorbate_species)
        self.descriptor = 'SimpleCutSpliceCrossoverWithAdsorbates'

        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.keep_composition = keep_composition
        self.fix_coverage = fix_coverage
        self.min_adsorbate_distance = min_adsorbate_distance
        self.adsorption_sites = adsorption_sites
        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': False, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.min_inputs = 2
        self.dmax = dmax
        
    def get_new_individual(self, parents):
        f, m = parents        
        indi = f.copy()

        if self.fix_coverage:
            # Count number of adsorbates
            adsorbates_in_parents = len(self.get_all_adsorbate_indices(f))

        theta = random.random() * 2 * np.pi  # 0,2pi
        phi = random.random() * np.pi  # 0,pi
        e = np.array((np.sin(phi) * np.cos(theta),
                      np.sin(theta) * np.sin(phi),
                      np.cos(phi)))
        eps = 0.0001
        
        f.translate(-f.get_center_of_mass())
        m.translate(-m.get_center_of_mass())
        
        # Get the signed distance to the cutting plane
        # We want one side from f and the other side from m
        mids, fids, rmids = [], [], []
        for i, x in enumerate(f.get_positions()):
            if np.dot(x, e) > 0:
                if f[i].symbol in adsorbate_elements:
                    rmids.append(i)
                else:
                    mids.append(i)
            else:
                if f[i].symbol not in adsorbate_elements:
                    fids.append(i)

        # Change half of f symbols to the half of m symbols
        for i in mids:
            indi[i].symbol = m[i].symbol

        # Check that the correct composition is employed
        if self.keep_composition:
            opt_sm = sorted([a.number for a in f if a.symbol 
                             not in adsorbate_elements])
            tmpf_numbers = list(indi.numbers[fids])
            tmpm_numbers = list(indi.numbers[mids])
            cur_sm = sorted(tmpf_numbers + tmpm_numbers)
            # correct_by: dictionary that specifies how many
            # of the atom_numbers should be removed (a negative number)
            # or added (a positive number)
            correct_by = dict([(j, opt_sm.count(j)) for j in set(opt_sm)])
            for n in cur_sm:
                correct_by[n] -= 1
            correct_ids = random.choice([fids, mids])
            to_add, to_rem = [], []
            for num, amount in correct_by.items():
                if amount > 0:
                    to_add.extend([num] * amount)
                elif amount < 0:
                    to_rem.extend([num] * abs(amount))
            for add, rem in zip(to_add, to_rem):
                tbc = [i for i in correct_ids if indi[i].number == rem]
                if len(tbc) == 0:
                    pass
                ai = random.choice(tbc)
                indi[ai].number = add

        # Place adsorbates from half of m and remove adsorbates from
        # half of f
        if self.adsorption_sites is not None:
            cas = deepcopy(self.adsorption_sites)
            cas.update(indi)
        else:
            cas = ClusterAdsorptionSites(indi, **self.kwargs) 
        fcac = ClusterAdsorbateCoverage(indi, cas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        fhsl = fcac.hetero_site_list
        mcac = ClusterAdsorbateCoverage(m, cas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        mhsl = mcac.hetero_site_list

        rmset = set(rmids)
        rmsites = []
        for st in fhsl:
            if st['occupied']:
                if not set(st['fragment_indices']).isdisjoint(rmset):
                    rmsites.append(st)

        # Remove fragments if every fragment is one of the given adsorbate species
        rmfrags = all(st['fragment'] in self.adsorbate_species for st in rmsites)
        remove_adsorbates_from_sites(indi, sites=rmsites, remove_fragments=rmfrags) 

        mset = set(mids)
        adsi_dict = {}
        for st in mhsl:
            if st['occupied']:
                si = st['indices']
                if set(si).issubset(mset):
                    si = st['indices']
                    adsi_dict[si] = {}
                    adsi_dict[si]['height'] = self.heights[st['site']]
                    adsi_dict[si]['fragment'] = st['fragment']
                    adsi_dict[si]['fragment_indices'] = st['fragment_indices']        

        nori = len(indi) 
        for st in cas.site_list:
            si = st['indices']
            if si in adsi_dict:
                adsorbate = adsi_dict[si]['fragment']
                height = adsi_dict[si]['height']    
                add_adsorbate_to_site(indi, adsorbate, st, height)

                # Make sure no adsorbates too close to each other 
                # after each adsorbate addition
                nads = len(adsi_dict[si]['fragment_indices'])
                if atoms_too_close_after_addition(indi, nads,
                self.min_adsorbate_distance, mic=False):
                    indi = indi[:-nads]                               

        # Add adsorbate if no adsorbate is present
        if len(indi) == nori:
            st = random.choice(cas.site_list)
            ads_spec = random.choice(self.adsorbate_species)
            add_adsorbate_to_site(indi, ads_spec, site=st, 
                                  height=self.heights[st['site']])

        cac = ClusterAdsorbateCoverage(indi, cas, subtract_heights=
                                       self.subtract_heights, dmax=self.dmax)
        ads_sites = cac.hetero_site_list

        if self.fix_coverage:
            # Remove or add adsorbates as needed
            adsorbates_in_child = self.get_all_adsorbate_indices(indi)
            diff = len(adsorbates_in_child) - adsorbates_in_parents
            if diff < 0:
                # Add adsorbates
                for _ in range(abs(diff)):
                    self.add_adsorbate(indi, ads_sites, site_heights,
                                       self.adsorbate_species,
                                       self.min_adsorbate_distance)
            elif diff > 0:
                # Remove adsorbates
                tbr = random.sample(adsorbates_in_child, diff)  # to be removed
                for adsorbate_indices in sorted(tbr, reverse=True):
                    for i in adsorbate_indices[::-1]:
                        indi.pop(i)                
            cac = ClusterAdsorbateCoverage(indi, cas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax)

        indi = self.initialize_individual(f, indi)
        indi.info['data']['parents'] = [i.info['confid'] for i in parents] 
        indi.info['data']['operation'] = 'crossover'
        indi.info['data']['adsorbates'] = [t[0] for t in 
            cac.get_adsorbates(self.adsorbate_species)]
        parent_message = ':Parents {0} {1}'.format(f.info['confid'],
                                                   m.info['confid'])
        return (self.finalize_individual(indi),
                self.descriptor + parent_message)

    def get_numbers(self, atoms):
        """Returns the atomic numbers of the atoms object
        without adsorbates"""
        ac = atoms.copy()
        del ac[[a.index for a in ac
                if a.symbol in adsorbate_elements]]
        return ac.numbers

    def get_atoms_without_adsorbates(self, atoms):
        ac = atoms.copy()
        del ac[[a.index for a in ac
                if a.symbol in adsorbate_elements]]
        return ac


class CatalystAdsorbateCrossover(AdsorbateOperator):                                   
    """Crossover that divides two particles or two slabs by the catalyst-
    adsorbates interfaces and exchange all adsorbates (only returns one 
    of them). The indexing of the atoms is preserved. Please only use 
    this operator with other operators that also preserves the indexing.

    The composition or the coverage is fixed if it is preserved by all 
    other operators being used.

    Parameters
    ----------
    group_by_sites : bool, default False
        If atoms.info['data']['groups'] is used, please set this to True
        if the groups is for adsorption sites, so that the offspring will
        inheritate the site groups. The default is grouping by slab atoms.

    """

    def __init__(self, group_by_sites=False):
        AdsorbateOperator.__init__(self, adsorbate_species='X')
        self.descriptor = 'CatalystAdsorbateCrossover'

        self.group_by_sites = group_by_sites
        self.min_inputs = 2
        
    def get_new_individual(self, parents):
        f, m = parents        
        indi = f.copy()
        slab_ids = [a.index for a in indi if a.symbol not in adsorbate_elements]
        indi.symbols[slab_ids] = m.symbols[slab_ids]

        indi = self.initialize_individual(f, indi)
        if 'groups' in f.info['data']:
            if self.group_by_sites:
                indi.info['data']['groups'] = f.info['data']['groups']
            else:
                indi.info['data']['groups'] = m.info['data']['groups']
        indi.info['data']['parents'] = [i.info['confid'] for i in parents] 
        indi.info['data']['operation'] = 'crossover'
        parent_message = ':Parents {0} {1}'.format(f.info['confid'],
                                                   m.info['confid'])
        return (self.finalize_individual(indi),
                self.descriptor + parent_message)




from acat.adsorption_sites import SlabAdsorptionSites
from acat.adsorbate_coverage import SlabAdsorbateCoverage
from acat.build.action import add_adsorbate
from ase.build import fcc211

atoms = fcc211('Cu', (3, 3, 4), vacuum=5.)
for atom in atoms:
    if atom.index % 2 == 0:
        atom.symbol = 'Au'
atoms.center()
add_adsorbate(atoms, adsorbate='CH3OH', surface='fcc211',
              indices=(5, 7, 8), surrogate_metal='Cu')
sac = SlabAdsorbateCoverage(atoms, surface='fcc211', 
                            label_occupied_sites=True)
occupied_sites = sac.get_sites(occupied_only=True)
assert len(occupied_sites) == 2




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import numpy as np
from operator import itemgetter

from ase.ga.offspring_creator import OffspringCreator
from ase.ga.utilities import get_distance_matrix, get_nndist
from ase import Atoms


class Mutation(OffspringCreator):
    """Base class for all particle mutation type operators.
    Do not call this class directly."""

    def __init__(self, num_muts=1):
        OffspringCreator.__init__(self, num_muts=num_muts)
        self.descriptor = 'Mutation'
        self.min_inputs = 1

    @classmethod
    def get_atomic_configuration(cls, atoms, elements=None, eps=4e-2):
        """Returns the atomic configuration of the particle as a list of
        lists. Each list contain the indices of the atoms sitting at the
        same distance from the geometrical center of the particle. Highly
        symmetrical particles will often have many atoms in each shell.

        For further elaboration see:
        J. Montejano-Carrizales and J. Moran-Lopez, Geometrical
        characteristics of compact nanoclusters, Nanostruct. Mater., 1,
        5, 397-409 (1992)

        Parameters:

        elements: Only take into account the elements specified in this
            list. Default is to take all elements into account.

        eps: The distance allowed to separate elements within each shell."""
        atoms = atoms.copy()
        if elements is None:
            e = list(set(atoms.get_chemical_symbols()))
        else:
            e = elements
        atoms.set_constraint()
        atoms.center()
        geo_mid = np.array([(atoms.get_cell() / 2.)[i][i] for i in range(3)])
        dists = [(np.linalg.norm(geo_mid - atoms[i].position), i)
                 for i in range(len(atoms))]
        dists.sort(key=itemgetter(0))
        atomic_conf = []
        old_dist = -10.
        for dist, i in dists:
            if abs(dist - old_dist) > eps:
                atomic_conf.append([i])
            else:
                atomic_conf[-1].append(i)
            old_dist = dist
        sorted_elems = sorted(set(atoms.get_chemical_symbols()))
        if e is not None and sorted(e) != sorted_elems:
            for shell in atomic_conf:
                torem = []
                for i in shell:
                    if atoms[i].symbol not in e:
                        torem.append(i)
                for i in torem:
                    shell.remove(i)
        return atomic_conf

    @classmethod
    def get_list_of_possible_permutations(cls, atoms, l1, l2):
        """Returns a list of available permutations from the two
        lists of indices, l1 and l2. Checking that identical elements
        are not permuted."""
        possible_permutations = []
        for i in l1:
            for j in l2:
                if atoms[int(i)].symbol != atoms[int(j)].symbol:
                    possible_permutations.append((i, j))
        return possible_permutations


class RandomMutation(Mutation):
    """Moves a random atom the supplied length in a random direction."""

    def __init__(self, length=2., num_muts=1, allowed_indices=None):
        Mutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'RandomMutation'
        self.length = length
        self.allowed_indices = allowed_indices

    def mutate(self, atoms):
        """ Does the actual mutation. """
        tbm = random.choice(range(len(atoms)))

        indi = Atoms()
        for a in atoms:
            if a.index == tbm:
                a.position += self.random_vector(self.length)
            indi.append(a)
        return indi

    def get_new_individual(self, parents):
        f0 = parents[0]
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        to_mut = f.copy()
        for _ in range(self.num_muts):
            to_mut = self.mutate(to_mut)

        for atom in to_mut:
            indi.append(atom)

        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def random_vector(cls, l):
        """return random vector of length l"""
        vec = np.array([random.random() * 2 - 1 for i in range(3)])
        vl = np.linalg.norm(vec)
        return np.array([v * l / vl for v in vec])


class RandomPermutation(Mutation):
    """Permutes two random atoms.

    Parameters:

    num_muts: the number of times to perform this operation.

    allowed_indices: The indices of atoms that are allowed to 
        be mutated by this operator. All indices are considered 
        if this is not specified."""

    def __init__(self, elements=None, num_muts=1, allowed_indices=None):
        Mutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'RandomPermutation'
        self.elements = elements
        self.allowed_indices = allowed_indices

    def get_new_individual(self, parents):
        f0 = parents[0].copy()
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        diffatoms = len(set(f.numbers))
        assert diffatoms > 1, 'Permutations with one atomic type is not valid'

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for _ in range(self.num_muts):
            RandomPermutation.mutate(f, self.elements)

        for atom in f:
            indi.append(atom)

        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def mutate(cls, atoms, elements=None):
        """Do the actual permutation."""
        if elements is None:
            indices = range(len(atoms))
        else:
            indices = [a.index for a in atoms if a.symbol in elements]
        i1 = random.choice(indices)
        i2 = random.choice(indices)
        while atoms[i1].symbol == atoms[i2].symbol:
            i2 = random.choice(indices)
        atoms.symbols[[i1, i2]] = atoms.symbols[[i2, i1]]


class COM2surfPermutation(Mutation):
    """The Center Of Mass to surface (COM2surf) permutation operator
    described in
    S. Lysgaard et al., Top. Catal., 2014, 57 (1-4), pp 33-39

    Parameters:

    elements: which elements should be included in this permutation,
        for example: include all metals and exclude all adsorbates

    min_ratio: minimum ratio of each element in the core or surface region.
        If elements=[a, b] then ratio of a is Na / (Na + Nb) (N: Number of).
        If less than minimum ratio is present in the core, the region defining
        the core will be extended until the minimum ratio is met, and vice
        versa for the surface region. It has the potential reach the
        recursive limit if an element has a smaller total ratio in the
        complete particle. In that case remember to decrease this min_ratio.

    num_muts: the number of times to perform this operation.

    allowed_indices: The indices of atoms that are allowed to 
        be mutated by this operator. All indices are considered 
        if this is not specified.
    """

    def __init__(self, elements=None, min_ratio=0.25, 
                 num_muts=1, allowed_indices=None):
        Mutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'COM2surfPermutation'
        self.min_ratio = min_ratio
        self.elements = elements
        self.allowed_indices = allowed_indices

    def get_new_individual(self, parents):
        f0 = parents[0].copy()
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        diffatoms = len(set(f.numbers))
        assert diffatoms > 1, 'Permutations with one atomic type is not valid'

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for _ in range(self.num_muts):
            elems = self.elements
            COM2surfPermutation.mutate(f, elems, self.min_ratio)

        for atom in f:
            indi.append(atom)

        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def mutate(cls, atoms, elements, min_ratio):
        """Performs the COM2surf permutation."""
        ac = atoms.copy()
        if elements is not None:
            del ac[[a.index for a in ac if a.symbol not in elements]]
        syms = ac.get_chemical_symbols()
        for el in set(syms):
            assert syms.count(el) / float(len(syms)) > min_ratio

        atomic_conf = Mutation.get_atomic_configuration(atoms,
                                                        elements=elements)
        core = COM2surfPermutation.get_core_indices(atoms,
                                                    atomic_conf,
                                                    min_ratio)
        shell = COM2surfPermutation.get_shell_indices(atoms,
                                                      atomic_conf,
                                                      min_ratio)
        permuts = Mutation.get_list_of_possible_permutations(atoms,
                                                             core,
                                                             shell)
        swap = list(random.choice(permuts))
        atoms.symbols[swap] = atoms.symbols[swap[::-1]]

    @classmethod
    def get_core_indices(cls, atoms, atomic_conf, min_ratio, recurs=0):
        """Recursive function that returns the indices in the core subject to
        the min_ratio constraint. The indices are found from the supplied
        atomic configuration."""
        elements = list(set([atoms[i].symbol
                             for subl in atomic_conf for i in subl]))

        core = [i for subl in atomic_conf[:1 + recurs] for i in subl]
        while len(core) < 1:
            recurs += 1
            core = [i for subl in atomic_conf[:1 + recurs] for i in subl]

        for elem in elements:
            ratio = len([i for i in core
                         if atoms[i].symbol == elem]) / float(len(core))
            if ratio < min_ratio:
                return COM2surfPermutation.get_core_indices(atoms,
                                                            atomic_conf,
                                                            min_ratio,
                                                            recurs + 1)
        return core

    @classmethod
    def get_shell_indices(cls, atoms, atomic_conf, min_ratio, recurs=0):
        """Recursive function that returns the indices in the surface
        subject to the min_ratio constraint. The indices are found from
        the supplied atomic configuration."""
        elements = list(set([atoms[i].symbol
                             for subl in atomic_conf for i in subl]))

        shell = [i for subl in atomic_conf[-1 - recurs:] for i in subl]
        while len(shell) < 1:
            recurs += 1
            shell = [i for subl in atomic_conf[-1 - recurs:] for i in subl]

        for elem in elements:
            ratio = len([i for i in shell
                         if atoms[i].symbol == elem]) / float(len(shell))
            if ratio < min_ratio:
                return COM2surfPermutation.get_shell_indices(atoms,
                                                             atomic_conf,
                                                             min_ratio,
                                                             recurs + 1)
        return shell


class _NeighborhoodPermutation(Mutation):
    """Helper class that holds common functions to all permutations
    that look at the neighborhoods of each atoms."""
    @classmethod
    def get_possible_poor2rich_permutations(cls, atoms, inverse=False,
                                            recurs=0, distance_matrix=None):
        dm = distance_matrix
        if dm is None:
            dm = get_distance_matrix(atoms)
        # Adding a small value (0.2) to overcome slight variations
        # in the average bond length
        nndist = get_nndist(atoms, dm) + 0.2
        same_neighbors = {}

        def f(x):
            return x[1]
        for i, atom in enumerate(atoms):
            same_neighbors[i] = 0
            neighbors = [j for j in range(len(dm[i])) if dm[i][j] < nndist]
            for n in neighbors:
                if atoms[n].symbol == atom.symbol:
                    same_neighbors[i] += 1
        sorted_same = sorted(same_neighbors.items(), key=f)
        if inverse:
            sorted_same.reverse()
        poor_indices = [j[0] for j in sorted_same
                        if abs(j[1] - sorted_same[0][1]) <= recurs]
        rich_indices = [j[0] for j in sorted_same
                        if abs(j[1] - sorted_same[-1][1]) <= recurs]
        permuts = Mutation.get_list_of_possible_permutations(atoms,
                                                             poor_indices,
                                                             rich_indices)

        if len(permuts) == 0:
            _NP = _NeighborhoodPermutation
            return _NP.get_possible_poor2rich_permutations(atoms, inverse,
                                                           recurs + 1, dm)
        return permuts


class Poor2richPermutation(_NeighborhoodPermutation):
    """The poor to rich (Poor2rich) permutation operator described in
    S. Lysgaard et al., Top. Catal., 2014, 57 (1-4), pp 33-39

    Permutes two atoms from regions short of the same elements, to
    regions rich in the same elements.
    (Inverse of Rich2poorPermutation)

    Parameters:

    elements: Which elements to take into account in this permutation

    num_muts: the number of times to perform this operation.

    allowed_indices: The indices of atoms that are allowed to 
        be mutated by this operator. All indices are considered 
        if this is not specified.
    """

    def __init__(self, elements=[], num_muts=1, allowed_indices=None):
        _NeighborhoodPermutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'Poor2richPermutation'
        self.elements = elements
        self.allowed_indices = allowed_indices

    def get_new_individual(self, parents):
        f0 = parents[0].copy()
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        diffatoms = len(set(f.numbers))
        assert diffatoms > 1, 'Permutations with one atomic type is not valid'

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        for _ in range(self.num_muts):
            Poor2richPermutation.mutate(f, self.elements)

        for atom in f:
            indi.append(atom)

        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def mutate(cls, atoms, elements):
        _NP = _NeighborhoodPermutation
        # indices = [a.index for a in atoms if a.symbol in elements]
        ac = atoms.copy()
        del ac[[atom.index for atom in ac
                if atom.symbol not in elements]]
        permuts = _NP.get_possible_poor2rich_permutations(ac)
        swap = list(random.choice(permuts))
        atoms.symbols[swap] = atoms.symbols[swap[::-1]]


class Rich2poorPermutation(_NeighborhoodPermutation):
    """
    The rich to poor (Rich2poor) permutation operator described in
    S. Lysgaard et al., Top. Catal., 2014, 57 (1-4), pp 33-39

    Permutes two atoms from regions rich in the same elements, to
    regions short of the same elements.
    (Inverse of Poor2richPermutation)

    Parameters:

    elements: Which elements to take into account in this permutation

    num_muts: the number of times to perform this operation.

    allowed_indices: The indices of atoms that are allowed to 
        be mutated by this operator. All indices are considered 
        if this is not specified.
    """

    def __init__(self, elements=None, num_muts=1, allowed_indices=None):
        _NeighborhoodPermutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'Rich2poorPermutation'
        self.elements = elements
        self.allowed_indices = allowed_indices

    def get_new_individual(self, parents):
        f0 = parents[0].copy()
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        diffatoms = len(set(f.numbers))
        assert diffatoms > 1, 'Permutations with one atomic type is not valid'

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        if self.elements is None:
            elems = list(set(f.get_chemical_symbols()))
        else:
            elems = self.elements
        for _ in range(self.num_muts):
            Rich2poorPermutation.mutate(f, elems)

        for atom in f:
            indi.append(atom)

        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def mutate(cls, atoms, elements):
        _NP = _NeighborhoodPermutation
        ac = atoms.copy()
        del ac[[atom.index for atom in ac
                if atom.symbol not in elements]]
        permuts = _NP.get_possible_poor2rich_permutations(ac,
                                                          inverse=True)
        swap = list(random.choice(permuts))
        atoms.symbols[swap] = atoms.symbols[swap[::-1]]


class SymmetricSubstitute(Mutation):
    """Permute all atoms within a subshell of the symmetric particle.
    The atoms within a subshell all have the same distance to the center,
    these are all equivalent under the particle point group symmetry.

    """

    def __init__(self, elements=None, num_muts=1, allowed_indices=None):
        Mutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'SymmetricSubstitute'
        self.elements = elements
        self.allowed_indices = allowed_indices

    def substitute(self, atoms):
        """Does the actual substitution"""
        atoms = atoms.copy()
        aconf = self.get_atomic_configuration(atoms,
                                              elements=self.elements)
        itbm = random.randint(0, len(aconf) - 1)
        to_element = random.choice(self.elements)

        for i in aconf[itbm]:
            atoms[i].symbol = to_element

        return atoms

    def get_new_individual(self, parents):
        f0 = parents[0]
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        indi = self.substitute(f)
        indi = self.initialize_individual(f, indi)
        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        indi.info['data']['parents'] = [f.info['confid']]

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))


class RandomSubstitute(Mutation):
    """Substitutes one atom with another atom type. The possible atom types
    are supplied in the parameter elements"""

    def __init__(self, elements=None, num_muts=1, allowed_indices=None):
        Mutation.__init__(self, num_muts=num_muts)
        self.descriptor = 'RandomSubstitute'
        self.elements = elements
        self.allowed_indices = allowed_indices

    def substitute(self, atoms):
        """Does the actual substitution"""
        atoms = atoms.copy()
        if self.elements is None:
            elems = list(set(atoms.get_chemical_symbols()))
        else:
            elems = self.elements[:]
        possible_indices = [a.index for a in atoms
                            if a.symbol in elems]
        itbm = random.choice(possible_indices)
        elems.remove(atoms[itbm].symbol)
        new_symbol = random.choice(elems)
        atoms[itbm].symbol = new_symbol

        return atoms

    def get_new_individual(self, parents):
        f0 = parents[0]
        if self.allowed_indices is None: 
            f = f0
        else:
            f = f0[self.allowed_indices]

        indi = self.substitute(f)
        indi = self.initialize_individual(f, indi)
        if self.allowed_indices is not None:
            tmp = f0.copy()
            tmp.symbols[self.allowed_indices] = indi.symbols
            info = indi.info.copy()
            indi = tmp.copy()
            indi.info = info

        indi.info['data']['parents'] = [f.info['confid']]

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))



from acat.build import add_adsorbate
from ase.build import fcc211
from ase.visualize import view

# Initialize a pure Cu(211) slab                             
atoms = fcc211('Cu', (3, 3, 4), vacuum=5.)

# Generate a CuAu(211) slab by changing some Cu to Au 
for atom in atoms:
    if atom.index % 2 == 0:
        atom.symbol = 'Au'

# Add a CO adsorbate to one site that is on the step bridge site
# with a composition of 'AuCu' and a height of 1.8 Angstrom
# (Always remember to pass in the surrogate metal for robust
# asorption site identification)
add_adsorbate(atoms, adsorbate='CO', site='bridge', surface='fcc211',
              morphology='step', height=1.8, surrogate_metal='Cu')

# Visualize the structure
view(atoms)

# Add another CO adsorbate to the site with atomic indices of (5,7,8)
add_adsorbate(atoms, adsorbate='CO', surface='fcc211', indices=(5,7,8), 
              height=1.8, surrogate_metal='Cu')

# Visualize the structure
view(atoms)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from ase.formula import Formula
from ase.build import molecule
from ase.parallel import parprint
from ase import Atoms
import numpy as np


# Adsorbate elements must be different from catalyst elements
adsorbate_elements = ['H', 'C', 'N', 'O', 'F', 'S', 'Cl']

# Adsorbate height on different sites
site_heights = {'ontop': 1.8, 
                'bridge': 1.5, 
                'shortbridge': 1.5,
                'longbridge': 1.5,
                'fcc': 1.3, 
                'hcp': 1.3,
                '3fold': 1.3, 
                '4fold': 1.3,
                '5fold': 1.5,
                '6fold': 0.,}

# The default adsorbate list already contains most common adsorbate 
# species. If you want to add any species, make sure you always sort 
# the indices of the atoms in the same order as the symbolm, by adding 
# entries in the adsorbate_molecule function. 

# Adsorbate nomenclature: first element always starts from bonded index 
# or the bonded element with smaller atomic number if multi-dentate.
# Hydrogen is not considered as a bonding element except for H and H2.
# The hydrogens should always follow the atom that they bond to.
# This is different from ASE's nomenclature, e.g. water should be 'OH2',
# hydrogen peroxide should be 'OHOH'

# Monodentate (vertical)                            
monodentate_adsorbate_list = ['H','C','N','O','S',
                              'CH','NH','OH','SH','CO','NO','CN','CS','NS',
                              'CH2','NH2','OH2','SH2','COH','NOH',
                              'CH3','NH3','OCH',
                              'OCH2',
                              'OCH3',]

                              # You may want to put some monodentate species here
# Multidentate (lateral)      # as potential multidentate species on rugged surfaces
multidentate_adsorbate_list = ['H2','C2','N2','O2','S2','OS',
                               'CO2','NO2','N2O','O2S','CS2','NS2','CHN','CHO','NHO','COS','C3','O3',
                               'CHOH','CH2O','COOH','CHOO','OHOH',
                               'CH3O','CH2OH','CH3S','CH2CO',
                               'CH3OH','CHOOH','CH3CO',
                               'CH3COOH',
                               'CHCHCHCHCHCH',]

adsorbate_list = monodentate_adsorbate_list + multidentate_adsorbate_list
adsorbate_formulas = {k: ''.join(list(Formula(k))) for k in adsorbate_list}

# Add entries and make your own adsorbate molecules
def adsorbate_molecule(adsorbate):
    # The ase.build.molecule module has many issues.       
    # Adjust positions, angles and indexing for your needs.
    if adsorbate == 'CO':
        ads = molecule(adsorbate)[::-1]
    elif adsorbate == 'C2':
        ads = molecule('C2H2')
        del ads[-2:]
        ads.rotate(90, 'x')
    elif adsorbate in ['H2','N2','O2','S2']:
        ads = molecule(adsorbate)
        ads.rotate(90, 'x')
    elif adsorbate == 'NS':
        ads = molecule('CS')
        ads[0].symbol = 'N'
    elif adsorbate == 'OS':
        ads = molecule('SO')
        ads.rotate(90, 'x')
    elif adsorbate == 'OH2':
        ads = molecule('H2O')
        ads.rotate(180, 'y')
    elif adsorbate == 'CH2':
        ads = molecule('NH2')
        ads[0].symbol = 'C'
        ads.rotate(180, 'y')
    elif adsorbate in ['NH2','SH2']:
        ads = molecule(adsorbate)
        ads.rotate(180, 'y')
    elif adsorbate == 'COH':
        ads = molecule('H2COH')
        del ads[-2:]
        ads.rotate(90, 'y')
    elif adsorbate == 'NOH':
        ads = molecule('H2COH')
        ads[0].symbol = 'N'
        del ads[-2:]
        ads.rotate(90, 'y')
    elif adsorbate == 'CO2':     
        ads = molecule(adsorbate)
        ads.rotate(90, 'x')
        ads.rotate(90, '-z')
    elif adsorbate in ['CS2','N2O']:
        ads = molecule(adsorbate)[[1,0,2]]
        ads.rotate(90, 'x')
    elif adsorbate == 'NS2':
        ads = molecule('CS2')[[1,0,2]]
        ads[0].symbol = 'N'
        ads.rotate(90, 'x')
    elif adsorbate == 'NO2':
        ads = molecule(adsorbate)
        ads.rotate(180, 'y')
    elif adsorbate == 'O2S':
        ads = molecule('SO2')[::-1]
        ads.rotate(180, 'y')
    elif adsorbate == 'CHN':
        ads = molecule('HCN')[[0,2,1]]
        ads.rotate(90, 'x')
        ads.rotate(90, '-z')
    elif adsorbate == 'CHO':
        ads = molecule('HCO')[[0,2,1]]
        ads.rotate(90, '-z')
    elif adsorbate == 'NHO':
        ads = molecule('HCO')[[0,2,1]]
        ads[0].symbol = 'N'
        ads.rotate(90, '-z')
    elif adsorbate == 'COS':
        ads = molecule('OCS')[[1,0,2]]
        ads.rotate(90, 'x')
    elif adsorbate == 'C3':
        ads = molecule('C3H4_D2d')
        del ads[-4:]
        ads.rotate(90, 'x')
    elif adsorbate == 'O3':
        ads = molecule(adsorbate)[[1,0,2]]
        ads.rotate(180, 'y')
    elif adsorbate == 'CH3':
        ads = molecule('CH3O')[[0,2,3,4]]
        ads.rotate(90, '-x')
    elif adsorbate == 'NH3':
        ads = molecule(adsorbate)
        ads.rotate(180, 'y')
    elif adsorbate == 'OCH2':
        ads = molecule('H2CO')
        ads.rotate(180, 'y')
    elif adsorbate == 'OCH3':
        ads = molecule('CH3O')[[1,0,2,3,4]]
        ads.rotate(90, '-x')
    elif adsorbate == 'CH2O':
        ads = molecule('H2CO')[[1,2,3,0]]
        ads.rotate(90, 'y')
        ads.rotate(180, '-z')
    elif adsorbate in ['CH3O','CH3S']:
        ads = molecule(adsorbate)[[0,2,3,4,1]]
        ads.rotate(30, 'y')
        ads.rotate(90, 'z')
    elif adsorbate == 'CH2CO':
        ads = molecule('H2CCO')[[0,2,3,1,4]]
        ads.rotate(90, 'y')
        ads.rotate(180, 'z')
    elif adsorbate == 'CHOH':
        ads = molecule('H2COH')
        del ads[-1]
        ads = ads[[0,3,1,2]]
    elif adsorbate == 'OHOH':
        ads = molecule('H2O2')[[0,2,1,3]]
        ads.rotate(90, '-z')
    elif adsorbate == 'CH2OH':
        ads = molecule('H2COH')[[0,3,4,1,2]]
    elif adsorbate == 'CH3OH':
        ads = molecule(adsorbate)[[0,2,4,5,1,3]]
        ads.rotate(-30, 'y')
        ads.rotate(90, '-z')
    elif adsorbate == 'CHOOH':
        ads = molecule('HCOOH')[[1,4,2,0,3]]
    elif adsorbate == 'CH3CO':
        ads = molecule(adsorbate)[[0,2,3,4,1,5]]
        ads.rotate(180, 'z')
    elif adsorbate == 'COOH':
        ads = Atoms('COOH', positions=[
                    (1.478221, 1.560435, 1.876707),
                    (2.124981, 0.480334, 2.067106),
                    (1.025178, 2.173870, 2.970078),
                    (0.550955, 3.006358, 2.725701),])
        ads.positions -= np.array([1.478221, 1.560435, 1.876707])
        ads.rotate(57, 'z')
    elif adsorbate == 'CHOO':
        ads = molecule('HCOOH')
        del ads[-2]
        ads = ads[[1,3,2,0]]
        ads.rotate(90, 'x')
        ads.rotate(7.5, 'y')
    elif adsorbate == 'CH3COOH':
        ads = molecule('CH3COOH')[[4,5,6,7,0,1,2,3]]
        ads.rotate(180, 'x')
    elif adsorbate == 'CHCHCHCHCHCH':
        ads = molecule('C6H6')[[0,6,1,7,2,8,3,9,4,10,5,11]] 
    elif adsorbate == 'X':
        ads = Atoms('X', positions=[[0, 0, 0]])
    else:
        try:
            ads = molecule(adsorbate)
        except:
            parprint('Molecule {} is not supported in the database.'.format(adsorbate))
            return 
    return ads




from acat.build.ordering import SymmetricClusterOrderingGenerator as SCOG
from ase.cluster import Octahedron
from ase.io import read
from ase.visualize import view

# Initialize a pure Ni truncated octahedral nanoparticle 
# as the template
atoms = Octahedron('Ni', 7, 2)

# Generate 50 symmetric nanoalloys with 3 metals with 
# "mirror circular" symmetry and save in a trajectory 
# file 'orderings_1.traj'
scog = SCOG(atoms, elements=['Ni', 'Cu', 'Pt'], 
            symmetry='mirror_circular',
            trajectory='orderings_1.traj')
scog.run(max_gen=50, mode='stochastic')

# Visualize the output structures
images1 = read('orderings_1.traj', index=':')
view(images1)

# Now generate 50 symmetric NiCuPt nanoalloys with 
# the composition of 1:1:1 and save in a trajectory 
# file 'orderings_2.traj'
scog = SCOG(atoms, elements=['Ni','Cu','Pt'], 
            symmetry='mirror_circular',
            composition={'Ni': 1,'Cu': 1,'Pt': 1}, 
            trajectory='orderings_2.traj')
scog.run(max_gen=50, mode='stochastic', eps=0.01)

# Visualize the output structures. 
images2 = read('orderings_2.traj', index=':')
view(images2)

